'''
Created on 2022-02-16

@author: wf
'''

class Version(object):
    '''
    Version handling for ConferenceCorpus
    '''
    version='0.0.35'
    date = '2020-09-10'
    updated = '2022-08-24'

        